package tp;

import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Locale;
import java.util.Set;
import java.util.stream.Collectors;

import org.graphstream.algorithm.APSP;
import org.graphstream.algorithm.APSP.APSPInfo;
import org.graphstream.algorithm.Dijkstra;
import org.graphstream.algorithm.Toolkit;
import org.graphstream.graph.Graph;
import org.graphstream.graph.Node;
import org.graphstream.graph.Path;
import org.graphstream.graph.implementations.SingleGraph;
import org.graphstream.stream.GraphParseException;

public class Main {

	private static final String RESOURCE_DIRECTORY = "./resources/";

	public static void main(String[] args) {
		Graph g = new SingleGraph("lh");
		
		try {
            g.read(RESOURCE_DIRECTORY + "lh.dgs");
        } catch (IOException e) {
            e.printStackTrace();
        } catch (GraphParseException e) {
            e.printStackTrace();
		}
        
		/*g.addAttribute("ui.stylesheet",
		"node { size: 3px; fill-color: #777; text-mode: hidden; z-index: 0;} " +
		"edge { shape: line; fill-color: #222; arrow-size: 3px, 2px;}");
		g.addAttribute("ui.quality");
		g.addAttribute("ui.antialias");

		g.display(false);*/

		//ALGORITHMS FEST
		
		/*System.out.println("Comparaison d'algorithmes de recherche de plus court chemin:\n");
		compareShortestPathAlgorithms(g);
		
		//DATA GATHERING
		
		int nodeCount = g.getNodeCount();
		int edgeCount = g.getEdgeCount();
		
		double avgDegree = Toolkit.averageDegree(g);
		double clusteringCoeff = Toolkit.averageClusteringCoefficient(g);
		int[] distribution = Toolkit.degreeDistribution(g);	
		
		//DATA PRINT
		
		System.out.format(
			Locale.FRENCH, "Mesures du graphe:\n\n\tNombre de noeuds: %d\n\tNombres d'arcs: %d\n\tDegré moyen: %.3f\n\tCoefficient de clustering: %.3f\n\n",
			nodeCount, edgeCount, avgDegree, clusteringCoeff
		);

		System.out.println("Distribution des degrés:");
		for(int i = 0; i < distribution.length; i++) {
			System.out.format(Locale.FRENCH, "\t%d: %d\n", i, distribution[i]);
		}
		System.out.println();*/

		/*Set<Node> centroid = findExcentricCenter(g);

		centroid.forEach(n -> n.addAttribute("ui.style", "fill-color: red;"));
		centroid.forEach(n -> n.addAttribute("ui.style", "size: 6px;"));*/

		displayBetweenessCentrality(g);
	}
	
	/**
	 * Gives the shortest path for two given nodes in a given graph using Dijkstra's algorithm
	 * @param graph The graph in which computes the algorithm
	 * @param source The source node for the shortest path
	 * @param dest The destination node for the shortest path
	 * @return The Path object representing the shortest path
	 */
	public static Path shortestPath(Graph graph, Node source, Node dest) {
		Dijkstra dAlgo = initDijkstra(graph, source);
		              
		return dAlgo.getPath(dest);
	}
	

	/**
	 * Computes the diameter of a given graph using the Dijkstra algorithm
	 * @param g The graph from which the diameter will be computed
	 * @return The Path object representing the shortest path
	 */
	public static Path computeDiameterDijkstra(Graph g) {
		Node source = null;
		Node dest = null;
		double longest = 0;
		
		for(Node n : g.getEachNode()) {
			Dijkstra dAlgo = initDijkstra(g, n);
			
			for(Node n2 : g.getEachNode()) {
				double wgt = dAlgo.getPathLength(n2);
				if(wgt > longest) {
					longest = wgt;
					source = n;
					dest = n2;
				}
			}
		}
		
		Path p = shortestPath(g, source, dest);
		
		System.out.println("\nDiametre: " + longest);

		return p;
	}
	

	/**
	 * Computes the diameter of a given graph using the Floyd-Warshall algorithm
	 * @param g The graph from which the diameter will be computed
	 * @return The Path object representing the shortest path
	 */
	public static Path computeDiameterFloydWarshall(Graph g) {
		Path path = null;
		double longest = 0;
		
		APSP apsp = new APSP();
        apsp.init(g); 
        apsp.setDirected(false); 
        apsp.setWeightAttributeName("length");
        apsp.compute();

        for(Node n : g.getEachNode()) {
        	APSPInfo info = n.getAttribute(APSPInfo.ATTRIBUTE_NAME);
        	
			for(Node n2 : g.getEachNode()) {
				Path p = info.getShortestPathTo(n2.getId());
				double wgt = p.getPathWeight("length");

				if(wgt > longest) {
					longest = wgt;
					path = p;
				}
			}
		}
        
		System.out.println("\nDiametre: " + longest);
		
		return path;
	}

	/**
	 * Prints an execution time comparison for shortest paths algorithms, specifically to find the diameter of a given graph
	 * @param g The graph from whixh the diameter will be computed
	 */
	public static void compareShortestPathAlgorithms(Graph g) {
		long startTime = System.currentTimeMillis();
		//computeDiameterFloydWarshall(g); 
		long endTime = System.currentTimeMillis();

		//System.out.println("Temps avec Floyd-Warshall: " + (endTime - startTime) + " millisecondes");
		
		startTime = System.currentTimeMillis();
		Path p = computeDiameterDijkstra(g);
		endTime = System.currentTimeMillis();
		
		System.out.println("Temps avec Djikstra: " + (endTime - startTime) + " millisecondes");

		p.getNodeSet().forEach(n -> n.addAttribute("ui.style", "fill-color: darkgreen;"));
		p.getEdgeSet().forEach(n -> n.addAttribute("ui.style", "fill-color: green;"));
	}

	/**
	 * Find the excentric center of a given graph
	 * @param g The graph from which the center is extracted
	 * @return a Set that represents a centroid (1 to n Node) for Graph g
	 */
	public static Set<Node> findExcentricCenter(Graph g) {
		Set<Node> centroid = new HashSet<>();
		Collection<Node> others = g.getNodeSet();
		int i = 0, nc = g.getNodeCount();

		int minExcentricity = nc;

		System.out.println("Debut du calcul du centre par excentricite");

		for (Node n : others) {
			int nodeExcentricity = 0;
			Dijkstra dAlgo = initDijkstra(g, n);

			for(Node n2 : others) {
				nodeExcentricity = Math.max(nodeExcentricity, dAlgo.getPath(n2).size());
			}

			n.setAttribute("ex", nodeExcentricity);
			minExcentricity = Math.min(minExcentricity, nodeExcentricity);

			System.out.format("\r%.2f %%", (double)++i / nc * 100);
		}

		System.out.println("\nFin du calcul");

		final int min = minExcentricity;

		others = g.getNodeSet();
		centroid.addAll(others.
						stream().
						filter(n -> (int)n.getAttribute("ex") == min).
						collect(Collectors.toSet()));

		return centroid;
	}


	/**
	 * Computes the betweeness centrality of each node in a given Graph
	 * @param g The Graph in which to compute the betweeness centrality of each node
	 */
	public static void computeBetweenessCentrality(Graph g) {
		Collection<Node> nodes = g.getNodeSet();

		//Init the bc attribute for every node
		nodes.forEach(n -> n.setAttribute("bc", 0));
		
		System.out.println("Debut du calcul du centre par intermediarite");

		int i = 0, nc = g.getNodeCount();
		for(Node n : nodes) {
			Dijkstra dAlgo = initDijkstra(g, n);
			
			for(Node n2 : nodes) {
				//for each node in the shortest path, increment their "hotness"
				dAlgo.getPath(n2).getNodeSet().forEach(
					no -> no.setAttribute("bc", (int)no.getAttribute("bc") + 1)
				);
			}

			System.out.format("\r%.2f %%", (double)++i / nc * 100);
		}

		System.out.println("\nFin du calcul");
		
		//we make a scale up to the max of the values to hav a range of values from 0 to 1
		double scale = g.getNodeSet().stream().map(n -> (int)n.getAttribute("bc")).reduce(0, Integer::max);
		
		//make each bc value of nodes to be a percentage 
		nodes.forEach(n -> n.setAttribute("bc", (int)n.getAttribute("bc") / scale));
	}


	/**
	 * Sets the ui and displays the graph with node color in function of betweeness centrality
	 */
	public static void displayBetweenessCentrality(Graph g) {
		computeBetweenessCentrality(g);

		g.addAttribute("ui.stylesheet",
		"node { size: 4px; fill-mode: dyn-plain; fill-color: blue, green, red; text-mode: hidden; z-index: 0;} " +
		"edge { shape: line; fill-color: #222; arrow-size: 3px, 2px;}");
		g.addAttribute("ui.quality");
		g.addAttribute("ui.antialias");

		g.getNodeSet().forEach(n -> n.setAttribute("ui.color", (double)n.getAttribute("bc")));
		
		//up the size of the node with the maximal betweeness centrality
		g.getNodeSet().stream().
						filter(n -> (double) n.getAttribute("bc") == g.getNodeSet().stream().
																					map(n2 -> (double)n2.getAttribute("bc")).
																					reduce(0.0, Double::max)
						).
						findFirst().
						get().setAttribute("ui.style", "size: 8px;");
	
		g.display(false);
	}


	/**
	 * Initialises a Dijkstra algorithm object on a given node
	 * @param g The Graph that contains the Node n
	 * @param n The Node to set the Dijkstra object on
	 * @return The Dijkstra object
	 */
	private static Dijkstra initDijkstra(Graph g, Node n) {
		Dijkstra d = new Dijkstra(Dijkstra.Element.EDGE, "result", "length");

		d.init(g);
		d.setSource(n);
		d.compute();

		return d;
	}
}
